import pandas as pd
import pytest

from siuba.dply.forcats import fct_reorder, fct_recode, fct_collapse, fct_lump, fct_inorder, fct_infreq
from pandas import Categorical

try:
    # pandas < v1
    from pandas.util.testing import assert_categorical_equal
except ImportError:
    from pandas._testing import assert_categorical_equal


def assert_fct_equal(x, y):
    return assert_categorical_equal(x,  y)

# fct_inorder -----------------------------------------------------------------

@pytest.mark.parametrize("x, dst_categories", [
    (["c", "a", "c", "b", "b"], ["c", "a", "b"]),
    (["c", "a", "c", "b", "b", None], ["c", "a", "b"])
])
def test_fct_inorder(x, dst_categories):
    dst = pd.Categorical(x, categories=dst_categories)

    res1 = fct_inorder(x)

    assert isinstance(res1, pd.Categorical)
    assert_categorical_equal(res1, dst)

    res2 = fct_inorder(pd.Series(x))

    assert isinstance(res2, pd.Series)
    assert_fct_equal(res2.array, dst)

    res3 = fct_inorder(pd.Categorical(x))

    assert isinstance(res3, pd.Categorical)
    assert_fct_equal(res3, dst)


@pytest.mark.parametrize("x, dst_categories", [
    (["c", "c", "b", "c", "a", "a"], ["c", "a", "b"]),              # no ties
    (["c", "c", "b", "c", "a", "a", "a"], ["c", "a", "b"]),          # ties
    (["c", "c", "b", "c", "a", "a", "a", None], ["c", "a", "b"])    # None
])
def test_fct_infreq(x, dst_categories):
    if pd.__version__.startswith("1.1.") or pd.__version__.startswith("1.2."):
        pytest.skip()

    dst = pd.Categorical(x, categories=dst_categories)

    res1 = fct_infreq(x)

    assert isinstance(res1, pd.Categorical)
    assert_categorical_equal(res1, dst)

    res2 = fct_infreq(pd.Series(x))

    assert isinstance(res2, pd.Series)
    assert_categorical_equal(res2.array, dst)

    res3 = fct_infreq(pd.Categorical(x))

    assert isinstance(res3, pd.Categorical)
    assert_categorical_equal(res3, dst)



# fct_reorder -----------------------------------------------------------------

def test_fct_reorder_simple():
    res = fct_reorder(['a', 'a', 'b'], [4, 3, 2])
    dst = Categorical(['a', 'a', 'b'], ['b', 'a'])

    assert_fct_equal(res, dst)
    
def test_fct_reorder_simple_upcast():
    res = fct_reorder(pd.Series(['a', 'a', 'b']), [4, 3, 2])
    dst = Categorical(['a', 'a', 'b'], ['b', 'a'])

    assert isinstance(res, pd.Series)
    assert_fct_equal(res.array, dst)

def test_fct_reorder_desc():
    res = fct_reorder(['a', 'a', 'b'], [4, 3, 2], desc = True)
    dst = Categorical(['a', 'a', 'b'], ['a', 'b'])

    assert_fct_equal(res, dst)

def test_fct_reorder_desc_upcast():
    res = fct_reorder(pd.Series(['a', 'a', 'b']), [4, 3, 2], desc = True)
    dst = Categorical(['a', 'a', 'b'], ['a', 'b'])

    assert isinstance(res, pd.Series)
    assert_fct_equal(res.array, dst)

def test_fct_reorder_custom_func():
    import numpy as np
    res = fct_reorder(['x', 'x', 'y'], [4, 0, 2], np.max)
    dst = Categorical(['x', 'x', 'y'], ['y', 'x'])

    assert_fct_equal(res, dst)

def test_fct_reorder_na_fct():
    import numpy as np
    res = fct_reorder([None, 'x', 'y'], [4, 3, 2], np.max)
    dst = Categorical([None, 'x', 'y'], ['y', 'x'])

    assert_fct_equal(res, dst)

# fct_recode ------------------------------------------------------------------

def test_fct_recode_simple():
    cat = ['a', 'b', 'c']
    res = fct_recode(cat, z = 'c')
    dst = Categorical(['a', 'b', 'z'], ['a', 'b', 'z'])

    assert_fct_equal(res, dst)

def test_fct_recode_simple_upcast():
    ser = pd.Series(['a', 'b', 'c'])
    res = fct_recode(ser, z = 'c')
    dst = Categorical(['a', 'b', 'z'], ['a', 'b', 'z'])

    assert isinstance(ser, pd.Series)
    assert_fct_equal(res.array, dst)


# fct_collapse ----------------------------------------------------------------

def test_fct_collapse_simple():
    res = fct_collapse(['a', 'b', 'c'], {'x': 'a'})
    dst = Categorical(['x', 'b', 'c'], ['x', 'b', 'c'])

    assert_fct_equal(res, dst)

def test_fct_collapse_simple_upcast():
    res = fct_collapse(pd.Series(['a', 'b', 'c']), {'x': 'a'})
    dst = Categorical(['x', 'b', 'c'], ['x', 'b', 'c'])

    assert isinstance(res, pd.Series)
    assert_fct_equal(res.array, dst)

def test_fct_collapse_others():
    res = fct_collapse(['a', 'b', 'c'], {'x': 'a'}, group_other = 'others')
    dst = Categorical(['x', 'others', 'others'], ['x', 'others'])

    assert_fct_equal(res, dst)

def test_fct_collapse_other_always_last():
    res = fct_collapse(['a', 'b', 'c'], {'x': 'c'}, group_other = 'others')
    dst = Categorical(['others', 'others', 'x'], ['x', 'others'])

    assert_fct_equal(res, dst)

def test_fct_collapse_many_to_one():
    res = fct_collapse(['a', 'b', 'c'], {'ab': ['a', 'b']})
    dst = Categorical(['ab', 'ab', 'c'], ['ab', 'c'])

    assert_fct_equal(res, dst)


# fct_lump --------------------------------------------------------------------

def test_fct_lump_n():
    res = fct_lump(['a', 'a', 'b', 'c'], n = 1)
    dst = Categorical(['a', 'a', 'Other', 'Other'], ['a', 'Other'])

    assert_fct_equal(res, dst)

def test_fct_lump_n_upcast():
    res = fct_lump(pd.Series(['a', 'a', 'b', 'c']), n = 1)
    dst = Categorical(['a', 'a', 'Other', 'Other'], ['a', 'Other'])

    assert isinstance(res, pd.Series)
    assert_fct_equal(res.array, dst)

def test_fct_lump_prop():
    res = fct_lump(['a', 'a', 'b', 'b', 'c', 'd'], prop = .2)
    dst = Categorical(['a', 'a', 'b', 'b', 'Other', 'Other'], ['a', 'b', 'Other'])

    assert_fct_equal(res, dst)



from siuba.dply.verbs import collect, show_query, mutate
from siuba.sql import LazyTbl
from siuba.dply.verbs import Pipeable
from siuba.tests.helpers import SqlBackend, data_frame
from siuba import _

import pandas as pd

import pytest

pg_backend = SqlBackend("postgresql")

@pytest.fixture(scope = "module")
def df_tiny():
    return pg_backend.load_df(data_frame(x = [1,2]))

@pytest.fixture(scope = "module")
def df_wide():
    return pg_backend.load_df(data_frame(x = [1,2], y = [3,4], z = [5, 6]))

def rename_source(query, tbl):
    return query.replace(tbl.tbl.name, "SRC_TBL")


def test_show_query_basic(df_tiny):
    q = df_tiny >> mutate(a = _.x.mean()) >> show_query(return_table = False)

    assert rename_source(str(q), df_tiny) == """\
SELECT SRC_TBL.x, avg(SRC_TBL.x) OVER () AS a 
FROM SRC_TBL"""

def test_show_query_basic_simplify(df_tiny):
    q = df_tiny >> mutate(a = _.x.mean()) >> show_query(return_table = False, simplify=True)

    assert rename_source(str(q), df_tiny) == """\
SELECT *, avg(SRC_TBL.x) OVER () AS a 
FROM SRC_TBL"""

def test_show_query_complex_simplify(df_wide):
    q = df_wide >>  mutate(a = _.x.mean(), b = _.a.mean())
    res = q >> show_query(return_table = False, simplify=True)

    assert rename_source(str(res), df_wide) == """\
SELECT *, avg(anon_1.a) OVER () AS b 
FROM (SELECT *, avg(SRC_TBL.x) OVER () AS a 
FROM SRC_TBL) AS anon_1"""




from . import (
    arrange,
    compute,
    conditional,
    count,
    distinct,
    explain,
    filter,
    group_by,
    head,
    join,
    mutate,
    select,
    summarize,
)

def __getattr__(name):
    import warnings

    if name == "LazyTbl":
        from ..backend import LazyTbl

        warnings.warn(
            "Importing LazyTbl from siuba.sql.verbs is deprecated. Please use "
            "`from siuba.sql import LazyTbl`",
            DeprecationWarning
        )
        return LazyTbl

    raise AttributeError(f"module 'siuba.sql.verbs' has no attribute '{name}'")




from siuba.siu import symbolic_dispatch
from types import SimpleNamespace

class Namespace(SimpleNamespace):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __iter__(self):
        return iter(self.__dict__)

    def __getitem__(self, x):
        return self.__dict__[x]

    def __len__(self):
        return len(self.__dict__)

    def keys(self):
        return self.__iter__()


def not_implemented(f):
    def _raise_not_impl(*args, **kwargs) -> NotImplementedError:
        raise NotImplementedError("Method %s not implemented" %f.__name__)

    return symbolic_dispatch(_raise_not_impl)

# register --------------------------------------------------------------------

def register(namespace, cls, **kwargs):
    """Register concrete methods on many dispatchers.

    Parameters:
        namespace: a mapping holding singledispatch functions
        cls: the class to being registered to
        kwargs: <name of generic> = <concrete method>

    """

    for k,v in kwargs.items():
        namespace[k].register(cls, v)

# Op --------------------------------------------------------------------------

class Operation:
    def __init__(self, name, kind, arity, is_property = False, accessor = None):
        self.name = name
        self.kind = kind
        self.arity = arity
        self.is_property = is_property
        self.accessor = accessor
    
def operation(name, *args):
    from siuba.siu.visitors import FunctionLookupBound
    # TODO: MC-NOTE - move FunctionLookupBound to dispatchers?

    msg = "No default singledispatch implemented for %s" % name
    dispatcher = symbolic_dispatch(FunctionLookupBound(msg))
    dispatcher.operation = Operation(name, *args)
    dispatcher.__name__ = name

    return dispatcher


# Default dispatcher ----------------------------------------------------------

def _register_series_default(generic):
    import pandas as pd
    from functools import partial

    generic.register(pd.Series, partial(_default_pd_series, generic.operation))

def _default_pd_series(__op, self, *args, **kwargs):
    # Once we drop python 3.7 dependency, could make __op position only
    if __op.accessor is not None:
        method = getattr(getattr(self, __op.accessor), __op.name)
    else:
        method = getattr(self, __op.name)


    if __op.is_property:
        return method

    return method(*args, **kwargs)




from siuba import spread, gather, _
from pandas.testing import assert_frame_equal
import pandas as pd

import pytest

@pytest.fixture
def df():
    return pd.DataFrame({
        'id': [1, 1, 2],
        'm': ['a', 'b', 'b'],
        'v': [1,2,3]
        })

@pytest.fixture
def wide_df():
    return pd.DataFrame({
        'id': [1,2],
        'a': [1, None],
        'b': [2., 3.]
        })

@pytest.fixture
def df_no_drop():
    return pd.DataFrame({
        'id': [1,2,1,2],
        'm': ['a', 'a', 'b', 'b'],
        'v': [1, None, 2, 3]
        })

@pytest.mark.parametrize('key, value', [
    ("m", "v"),
    (_.m, _.v),
    (_["m"], _["v"]),
    (1, 2),
    ])
def test_spread_selection(df, wide_df, key, value):
    res = spread(df, key, value)


    assert_frame_equal(res, wide_df)
    

def test_spread_fill(df, wide_df):
    res = spread(df, "m", "v", fill = 99)

    assert_frame_equal(res, wide_df.fillna(99.))


def test_spread_grouped_df(df, wide_df):
    gdf = df.groupby('id')
    res = spread(gdf, "m", "v")

    gdf_wide = wide_df.groupby('id')
    assert_frame_equal(res.obj, gdf_wide.obj)

    assert len(res.grouper.groupings) == 1
    assert res.grouper.groupings[0].name == "id"


def test_gather(df, wide_df):
    res = gather(wide_df, "m", "v", "a", "b") 

    # note, unlike df, includes an NA now
    long = pd.DataFrame({
        'id': [1,2,1,2],
        'm': ['a', 'a', 'b', 'b'],
        'v': [1, None, 2, 3]
        })

    assert_frame_equal(res, long)


def test_gather_drop_na(df, wide_df):
    res = gather(wide_df, "m", "v", "a", "b", drop_na = True) 

    # note that .dropna doesn't work here, since coerces floats to ints
    assert_frame_equal(res, df.assign(v = df.v.astype(float)))


def test_gather_drop_na_varselect(df, wide_df):
    res = gather(wide_df, "m", "v", _["a": "b"], drop_na = True) 

    # note that .dropna doesn't work here, since coerces floats to ints
    assert_frame_equal(res, df.assign(v = df.v.astype(float)))


def test_gather_no_selection_gathers_all(wide_df):
    res_implicit = gather(wide_df, "m", "v")
    res_explicit = gather(wide_df, "m", "v", _[:])

    assert res_implicit.columns.tolist() == ["m", "v"]
    assert "id" in res_implicit["m"].values

    assert_frame_equal(res_implicit, res_explicit)


def test_gather_no_match_gathers_none(wide_df):
    res = gather(wide_df, "m", "v", _.startswith("zzzz"))

    assert_frame_equal(res, wide_df)


def test_gather_group_by_no_match_still_grouped(wide_df):
    wide_gdf = wide_df.groupby("id")
    res = gather(wide_gdf, "m", "v", _.startswith("zzzz"))

    assert res is wide_gdf


def test_gather_group_by_preserves_groups(df_no_drop, wide_df):
    res = gather(wide_df.groupby("id"), "m", "v", _.a, _.b)
    
    groupings = res.grouper.groupings
    
    assert len(groupings) == 1
    assert groupings[0].name == "id"

    assert_frame_equal(res.obj, df_no_drop)


def test_gather_group_by_drops_gathered_groups(df_no_drop, wide_df):
    g_res = gather(wide_df.groupby(["id", "a"]), "m", "v", _.a, _.b)

    groupings = g_res.grouper.groupings
    assert len(groupings) == 1
    assert groupings[0].name == "id"

    assert_frame_equal(g_res.obj, df_no_drop)


def test_gather_group_by_drops_all_groups(df_no_drop, wide_df):
    res = gather(wide_df.groupby(["a", "b"]), "m", "v", _.a, _.b)

    assert isinstance(res, pd.DataFrame)

    assert_frame_equal(res, df_no_drop)





from sqlalchemy.sql import func as fn
from sqlalchemy import sql

from ..translate import (
    SqlTranslator,
    extend_base,
    sql_scalar,
    sql_agg,
    win_agg,
    win_cumul,
    annotate
)

#from .postgresql import PostgresqlColumn, PostgresqlColumnAgg
from .base import SqlColumn, SqlColumnAgg
from . import _dt_generics as _dt

# Data ----

class SnowflakeColumn(SqlColumn): pass
class SnowflakeColumnAgg(SqlColumnAgg, SnowflakeColumn): pass


# Translations ================================================================

@_dt.sql_func_last_day_in_period.register
def sql_func_last_day_in_period(codata: SnowflakeColumn, col, period):
    return _dt.date_trunc(codata, col, period) \
            + sql.text("interval '1 %s'" % period) \
            - sql.text("interval '1 day'")

# Scalar ----
extend_base(
    SnowflakeColumn,
    __floordiv__ = lambda _, x, y: fn.floor(x / y),
    __rfloordiv__ = lambda _, x, y: fn.floor(y / x),

    # connector has a bug with %
    # see: https://github.com/snowflakedb/snowflake-sqlalchemy/issues/246
    __mod__ = lambda _, x, y: fn.mod(x, y),
    __rmod__ = lambda _, x, y: fn.mod(y, x),
    mod = lambda _, x, y: fn.mod(x,y),
    rmod = lambda _, x, y: fn.mod(y,x),

    # TODO: str.contains
)

# Window ----

extend_base(
    SnowflakeColumn,
    all = win_agg("booland_agg"),
    any = win_agg("boolor_agg"),
    count = win_agg("count"),
    cumsum = annotate(win_cumul("sum"), result_type="variable"),
    # note that the number of decimal places Snowflake returns, and whether
    # the result is numeric depends on the input. mark as variable, so tests
    # do not check dtype
    # see https://community.snowflake.com/s/question/0D50Z000079hpxvSAA/numeric-calculations-truncated-to-3-decimal-places
    mean = annotate(win_agg("avg"), result_type="variable"),
    std = win_agg("stddev_samp"),
    sum = annotate(win_agg("sum"), result_type="variable"),

    var = win_agg("var_samp"),

    # str.contains
    # dt methods are more like base
)

# Agg ----

extend_base(
    SnowflakeColumnAgg,
    all = sql_agg("booland_agg"),
    any = sql_agg("boolor_agg"),
    count = sql_agg("count"),

    std = sql_agg("stddev_samp"),
    var = sql_agg("var_samp"),
)


translator = SqlTranslator.from_mappings(
        SnowflakeColumn, SnowflakeColumnAgg
        )



"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-group_by.R
"""
    
from siuba import _, group_by, ungroup, summarize, collect
from siuba.dply.vector import row_number, n
from siuba.dply import verbs

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, SqlBackend
from string import ascii_lowercase 

DATA = data_frame(x = [1,2,3], y = [9,8,7], g = ['a', 'a', 'b'])

@pytest.fixture(scope = "module")
def df(backend):
    if not isinstance(backend, SqlBackend):
        pytest.skip("TODO: generalize tests to pandas")
    return backend.load_df(DATA)


def test_group_by_two(df):
    gdf = group_by(df, _.x, _.y)
    assert gdf.group_by == ("x", "y")

def test_group_by_override(df):
    gdf = df >> group_by(_.x, _.y) >> group_by(_.g)
    assert gdf.group_by == ("g",)

def test_group_by_no_add(df):
    # without add argument, group_by overwrites prev grouping
    gdf1 = group_by(df, _.x)
    gdf2 = group_by(gdf1, _.y)

    assert gdf2.group_by == ("y",)

def test_group_by_add(df):
    gdf = group_by(df, _.x) >> group_by(_.y, add = True)

    assert gdf.group_by == ("x", "y")

def test_group_by_ungroup(df):
    q1 = df >> group_by(_.g)
    assert q1.group_by == ("g",)

    q2 = q1 >> ungroup()
    assert q2.group_by == tuple()

def test_group_by_using_string(df):
    gdf = group_by(df, "g") >> summarize(res = _.x.mean())
    

def test_group_by_performs_mutate(df):
    assert_equal_query(
            df,
            group_by(z = _.x + _.y) >> summarize(n = n(_)),
            data_frame(z = 10, n = 3)
            )


@pytest.mark.parametrize("query", [
    (verbs.rename(_, z = "y")),
    (verbs.mutate(_, z = _.x + 1)),
    (verbs.filter(_, _.x > 1)),
    (verbs.arrange(_, _.x)),
    (verbs.add_count(_, _.x)),
    (verbs.head(_)),
])
def test_group_by_equiv_ungrouped(backend, query):
    # TODO: check indexes
    df = data_frame(x = [1, 1, 2], y = [4, 5, 6], g = ["a", "a", "a"])
    src = backend.load_df(df)
    
    assert_equal_query(
        src,
        group_by(_.g) >> query,
        src >> query >> collect()
    )


@pytest.mark.parametrize("query", [
    (verbs.select(_, _.x, _.y)),
    (verbs.transmute(_, z = _.x + 1)),
    (verbs.summarize(_, avg = _.x.mean())),
    (verbs.count(_, _.x)),
    (verbs.distinct(_, _.x)),
])
def test_group_by_equiv_with_groupings(backend, query):
    df = data_frame(x = [1, 1, 2], y = [4, 5, 6], g = ["a", "a", "a"])
    src = backend.load_df(df)

    res = df >> query >> collect()
    res.insert(0, "g", "a")

    assert_equal_query(
        src,
        group_by(_.g) >> query,
        res
    )


@pytest.mark.parametrize("f_join", [
    (verbs.inner_join),
    (verbs.left_join),
    (verbs.right_join),
    (verbs.full_join),
    (verbs.semi_join),
    (verbs.anti_join),
])
def test_group_by_equiv_joins(xfail_backend, backend, f_join):
    if f_join is verbs.full_join and backend.name in {"sqlite", "mysql"}:
        pytest.xfail()

    df = data_frame(id = [1, 2, 3], val = ['lhs.1', 'lhs.2', 'lhs.3'])
    df2 = data_frame(id = [1, 2, 4], val = ['rhs.1', 'rhs.2', 'rhs.3'])

    src = backend.load_df(df)
    src2 = backend.load_df(df2)

    dst = src >> f_join(_, src2, by = "id") >> collect()

    assert_equal_query(
        src,
        group_by(_.id) >> f_join(_, src2, by = "id"),
        dst
    )
    




from siuba import _, group_by, separate, extract, unite

import pytest
from .helpers import assert_equal_query, assert_frame_sort_equal, data_frame, backend_notimpl, backend_sql

DATA = data_frame(label = ["S1-E1", "S1-E2"])

@pytest.fixture(scope = "function")
def df():
    return DATA.copy()

def test_separate_default(df):
    assert_equal_query(
            df,
            separate("label", into = ["season", "episode"]),
            data_frame(season = ["S1", "S1"], episode = ["E1", "E2"])
            )

def test_extract_default(df):
    assert_equal_query(
            df,
            extract("label", into = ["season"]),
            data_frame(season = ["S1", "S1"])
            )

def test_separate_into_replace_column():
    src = data_frame(x = [1, 2], label = ["S1-E1", "S1-E2"])
    dst = data_frame(x = ["S1", "S1"], episode = ["E1", "E2"])

    assert_equal_query(
            src,
            separate("label", into = ["x", "episode"]),
            dst
            )


def test_extract_into_replace_column():
    src = data_frame(x = [1, 2], label = ["S1-E1", "S1-E2"])
    dst = data_frame(x = ["S1", "S1"], episode = ["E1", "E2"])

    assert_equal_query(
            src,
            extract("label", into = ["x", "episode"], regex="(.*)?-(.*)"),
            dst
            )

# regexes ----
def test_separate_sep_arg(df):
    assert_equal_query(
            df,
            separate("label", into = ["season", "episode"], sep = "E"),
            data_frame(season = ["S1-", "S1-"], episode = ["1", "2"]),
            )

def test_extract_regex_arg(df):
    assert_equal_query(
            df,
            extract("label", into = ["season", "episode"], regex = "S([0-9]+)-E([0-9]+)"),
            data_frame(season = ["1", "1"], episode = ["1", "2"]),
            )

def test_extract_regex_arg_ignores_named(df):
    assert_equal_query(
            df,
            extract("label", into = ["season"], regex = "(?P<digit>[0-9]+)"),
            data_frame(season = ["1", "1"]),
            )


# conversions ----
def test_separate_convert_arg():
    data = data_frame(label = ["1-1", "2-a"])
    assert_equal_query(
            data,
            separate("label", into = ["season", "episode"], convert = True),
            data_frame(season = [1, 2], episode = ["1", "a"])
            )

def test_extract_convert_arg():
    data = data_frame(label = ["1-1", "2-a"])
    assert_equal_query(
            data,
            extract("label", into = ["season", "episode"], convert = True, regex = "(.)-(.)"),
            data_frame(season = [1, 2], episode = ["1", "a"])
            )

# misc ----
def test_separate_warn_arg_warn():
    data = data_frame(label = "1-2-3-4")
    with pytest.warns(UserWarning):
        separate(data, "label", into = ["a", "b"], sep = "-")

@pytest.mark.skip("TODO")
def test_separate_warn_arg_merge():
    pass

@pytest.mark.skip("TODO")
def test_separate_fill_arg():
    pass

def test_separate_remove_arg(df):
    assert_equal_query(
            df,
            separate("label", into = ["season", "episode"], remove = False),
            df.assign(season = ["S1", "S1"], episode = ["E1", "E2"])
            )

def test_extract_remove_arg(df):
    assert_equal_query(
            df,
            extract("label", into = ["season"], remove = False),
            df.assign(season = ["S1", "S1"])
            )

# unite =======================================================================

def test_unite_round_trip():
    data = data_frame(label = ["s1_e1", "s2_e2"])
    separated = separate(data, "label", into = ["season", "episode"])
    united = unite(separated, "label", "season", "episode")

    assert_frame_sort_equal(united, data)

def test_unite_round_trip_grouped_df():
    data = data_frame(label = ["s1_e1", "s2_e2"])
    separated = separate(data.groupby('label'), "label", into = ["season", "episode"])
    united = unite(separated, "label", "season", "episode")

    assert_frame_sort_equal(united, data)


def test_unite_missing_column_error():
    data = data_frame(a = ["a"], b = ["b"])

    with pytest.raises(ValueError):
        # TODO: should check message includes name of col
        united = unite(data, "label", "a", "c")




import re
import pandas as pd

from collections import Counter
from typing import Callable



def group_vars(df: pd.DataFrame) -> "list[str]":
    groupings = df.grouper.groupings
    group_cols = [ping.name for ping in groupings]
    return group_cols


def reconstruct_tibble(
    input_: pd.DataFrame,
    output: pd.DataFrame,
    ungrouped_vars: "tuple[str]" = tuple()
):
    if isinstance(input_, pd.core.groupby.DataFrameGroupBy):
        old_groups = group_vars(input_)
        new_groups = list((set(old_groups) - set(ungrouped_vars)) & set(output.columns))

        return output.groupby(new_groups)
    
    return output


def check_dict_of_functions(d_funcs: "Callable | dict[str, Callable]", names, arg_name):
    if isinstance(d_funcs, dict):
        missing = set(d_funcs) - set(names)
        if missing:
            raise ValueError(
                f"`{arg_name}=` is a dictionary of functions, so must corresponding to "
                "column names in the data. No matching column names found for these "
                f"entries: {missing}"
            )
        return d_funcs
    elif callable(d_funcs):
        return {k: d_funcs for k in names}

    raise TypeError(
        f"{arg_name} must be a dictionary mapping column names to functions, or "
        "a single function."
    )


# vec_as_names ----------------------------------------------------------------

SEPERATOR = "___"

def _make_unique(name):
    pass

def _strip_suffix(name):
    return re.sub("(___?[0-9]*)+$", "", name)

def vec_as_names(names, *, repair: "str | Callable"):
    """Validate and repair column names.

    Parameters
    ----------
    names:
        A list-like of column names
    repair:
 
    """
    # minimal, unique, universal, check_unique
    # minimal: names can be accessed using df[<name>], e.g. "x", np.nan
    # unique: names are minimal and no duplicates. Can be accessed using df[name]
    # check_unique: 
    # universal: accessible by attribute (may throw an error if its a pandas method)

    if repair not in {"unique", "check_unique", "minimal"} and not callable(repair):
        raise NotImplementedError()

    validate_unique = callable(repair) or repair == "check_unique"

    # minimal ---
    if repair == "minimal":
        return names

    # custom function ---
    if callable(repair):
        names = repair(names)

    # check_unique ----
    raw_uniq = Counter(names)
    if len(raw_uniq) < len(names) and validate_unique:
        duplicates = [entry for entry, n in raw_uniq.items() if n > 1]

        raise ValueError(
            f"Names must be unique, but detected {len(duplicates)} duplicate name(s).\n\n"
            f"Duplicated names: {duplicates}"
        )


    stripped_names = list(map(_strip_suffix, names))
    uniq = Counter(stripped_names)

    # name repair unique ----
    result = []
    for ii, name in enumerate(stripped_names):
        if uniq[name] > 1:
            result.append(f"{name}___{ii}")
        else:
            result.append(name)
    
    return result
    



from siuba.dply.verbs import group_by, ungroup

from ..backend import LazyTbl, ordered_union
from ..utils import lift_inner_cols

from .mutate import _mutate_cols


@group_by.register(LazyTbl)
def _group_by(__data, *args, add = False, **kwargs):
    if not (args or kwargs):
        return __data.copy()

    group_names, sel = _mutate_cols(__data, args, kwargs, "Group by")

    if None in group_names:
        raise NotImplementedError("Complex, unnamed expressions not supported in sql group_by")

    # check whether we can just use underlying table ----
    new_cols = lift_inner_cols(sel)
    if set(new_cols).issubset(set(__data.last_op.columns)):
        sel = __data.last_op

    if add:
        group_names = ordered_union(__data.group_by, group_names)

    return __data.append_op(sel, group_by = tuple(group_names))


@ungroup.register(LazyTbl)
def _ungroup(__data):
    return __data.copy(group_by = tuple())



import pandas as pd
from pandas.api import types as pd_types

from pandas.core.groupby import DataFrameGroupBy
from .verbs import var_select, var_create
from ..siu import FormulaContext, Call, strip_symbolic, Fx, FuncArg
from ..siu.dispatchers import verb_dispatch, symbolic_dispatch, create_eager_pipe_call

from collections.abc import Mapping
from contextvars import ContextVar
from contextlib import contextmanager
from typing import Callable, Any

DEFAULT_MULTI_FUNC_TEMPLATE = "{col}_{fn}"
DEFAULT_SINGLE_FUNC_TEMPLATE = "{col}"


ctx_verb_data = ContextVar("data")
ctx_verb_window = ContextVar("window")


def _is_symbolic_operator(f):
    # TODO: consolidate these checks, make the result of symbolic_dispatch a class.
    return callable(f) and getattr(f, "_siu_symbolic_operator", False)


def _require_across(call, verb_name):
    if (
        not isinstance(call, Call) 
        or not (call.args and getattr(call.args[0], "__name__", None) == "across")
    ):
        raise NotImplementedError(
            f"{verb_name} currently only allows a top-level across as an unnamed argument.\n\n"
            f"Example: {verb_name}(some_data, across(...))"
        )


def _eval_with_context(ctx, window_ctx, data, expr):
    # TODO: should just set the translator as context (e.g. agg translater, etc..)
    token = ctx_verb_data.set(ctx)
    token_win = ctx_verb_window.set(window_ctx)

    try:
        return expr(data)
    finally:
        ctx_verb_data.reset(token)
        ctx_verb_window.reset(token_win)


@contextmanager
def _set_data_context(ctx, window):
    try:
        token = ctx_verb_data.set(ctx)
        token_win = ctx_verb_window.set(window)
        yield
    finally:
        ctx_verb_data.reset(token)
        ctx_verb_window.reset(token_win)



# TODO: handle DataFrame manipulation in pandas / sql backends
class AcrossResult(Mapping):
    def __init__(self, *args, **kwargs):
        self.d = dict(*args, **kwargs)

    def __getitem__(self, k):
        return self.d[k]

    def __iter__(self):
        return iter(self.d)

    def __len__(self):
        return len(self.d)


def _across_setup_fns(fns) -> "dict[str, Callable[[FormulaContext], Any]]":
    final_calls = {}
    if isinstance(fns, (list, tuple)):
        raise NotImplementedError(
            "Specifying functions as a list or tuple is not supported. "
            "Please use a dictionary to define multiple functions to apply. \n\n"
            "E.g. across(_[:], {'round': Fx.round(), 'round2': Fx.round() + 1})"
        )
    elif isinstance(fns, dict):
        for name, fn_call_raw in fns.items():
            # symbolics get stripped by default for arguments to verbs, but
            # these are inside a dictionary, so need to strip manually.
            fn_call = strip_symbolic(fn_call_raw)

            if isinstance(fn_call, Call):
                final_calls[name] = fn_call

            elif callable(fn_call):
                final_calls[name] = create_eager_pipe_call(FuncArg(fn_call), Fx)

            else:
                raise TypeError(
                    "All functions to be applied in across must be a siuba.siu.Call, "
                    f"but received a function of type {type(fn_call)}"
                )

    elif isinstance(fns, Call):
        final_calls["fn1"] = fns

    elif callable(fns):
        final_calls["fn1"] = create_eager_pipe_call(FuncArg(fns), Fx)

    else:
        raise NotImplementedError(f"Unsupported function type in across: {type(fns)}")

    return final_calls


def _get_name_template(fns, names: "str | None") -> str:
    if names is not None:
        return names

    if callable(fns):
        return DEFAULT_SINGLE_FUNC_TEMPLATE

    return DEFAULT_MULTI_FUNC_TEMPLATE


@verb_dispatch(pd.DataFrame)
def across(__data, cols, fns, names: "str | None" = None) -> pd.DataFrame:

    name_template = _get_name_template(fns, names)
    selected_cols = var_select(__data.columns, *var_create(cols), data=__data)

    fns_map = _across_setup_fns(fns)

    results = {}
    for old_name, new_name in selected_cols.items():
        if new_name is None:
            new_name = old_name

        crnt_ser = __data[old_name]
        context = FormulaContext(Fx=crnt_ser, _=__data)

        for fn_name, fn in fns_map.items():
            fmt_pars = {"fn": fn_name, "col": new_name}
            
            res = fn(context)
            results[name_template.format(**fmt_pars)] = res

    # ensure at least one result is not a scalar, so we don't get the classic
    # pandas error: "If using all scalar values, you must pass an index"
    index = None
    if results:
        _, v = next(iter(results.items()))
        if pd_types.is_scalar(v):
           index = [0] 

    return pd.DataFrame(results, index=index)


@symbolic_dispatch(cls = pd.Series)
def where(x) -> bool:
    if not isinstance(x, bool):
        raise TypeError("Result of where clause must be a boolean (True or False).")

    return x
    



"""
This module allows you to check types (e.g. using isinstance) without importing them.

Note that this is copied from https://github.com/machow/databackend

"""

import sys
import importlib

from abc import ABCMeta


def _load_class(mod_name: str, cls_name: str):
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


class _AbstractBackendMeta(ABCMeta):
    def __new__(mcls, clsname, bases, attrs):
        cls = super().__new__(mcls, clsname, bases, attrs)
        if not hasattr(cls, "_backends"):
            cls._backends = []
        return cls

    def register_backend(cls, mod_name: str, cls_name: str):
        cls._backends.append((mod_name, cls_name))
        cls._abc_caches_clear()


class AbstractBackend(metaclass=_AbstractBackendMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        for mod_name, cls_name in cls._backends:
            if mod_name not in sys.modules:
                # module isn't loaded, so it can't be the subclass
                # we don't want to import the module to explicitly run the check
                # so skip here.
                continue
            else:
                target_cls = _load_class(mod_name, cls_name)
                if issubclass(subclass, target_cls):
                    return True

        return NotImplemented


# Implementations -------------------------------------------------------------

class SqlaEngine(AbstractBackend): pass
class PlDataFrame(AbstractBackend): pass
class PdDataFrame(AbstractBackend): pass

SqlaEngine.register_backend("sqlalchemy.engine", "Connectable")
PlDataFrame.register_backend("polars", "DataFrame")
PdDataFrame.register_backend("pandas", "DataFrame")



from siuba.dply.verbs import arrange, _call_strip_ascending
from siuba.dply.across import _set_data_context

from ..utils import lift_inner_cols
from ..backend import LazyTbl
from ..translate import ColumnCollection

# Helpers ---------------------------------------------------------------------

@arrange.register(LazyTbl)
def _arrange(__data, *args):
    # Note that SQL databases often do not subquery order by clauses. Arrange
    # sets order_by on the backend, so it can set order by in over elements,
    # and handle when new columns are named the same as order by vars.
    # see: https://dba.stackexchange.com/q/82930

    last_sel = __data.last_select
    cols = lift_inner_cols(last_sel)

    # TODO: implement across in arrange
    sort_cols = _eval_arrange_args(__data, args, cols)

    final_sel = last_sel.order_by(None).order_by(*sort_cols)
    return __data.append_op(final_sel, order_by = tuple(args))


def _eval_arrange_args(__data, args, cols):
    sort_cols = []
    for ii, expr in enumerate(args):
        shaped = __data.shape_call(
                expr, window = False, str_accessors = True,
                verb_name = "Arrange", arg_name = ii,
                )
        
        new_call, ascending = _call_strip_ascending(shaped)

        with _set_data_context(__data, window=True):
            res = new_call(cols)

        if isinstance(res, ColumnCollection):
            raise NotImplementedError(
                f"`arrange()` expression {ii} of {len(args)} returned multiple columns, "
                "which is currently unsupported."
            )

        if not ascending:
            res = res.desc()

        sort_cols.append(res)

    return sort_cols




from .backend import LazyTbl, sql_raw
from .translate import SqlColumn, SqlColumnAgg, SqlFunctionLookupError
from . import across as _across

# proceed w/ underscore so it isn't exported by default
# we just want to register the singledispatch funcs
from . import verbs as _verbs
from .dply import vector as _vector
from .dply import string as _string



from ..translate import SqlTranslator

from .postgresql import PostgresqlColumn, PostgresqlColumnAgg

class RedshiftColumn(PostgresqlColumn): pass
class RedshiftColumnAgg(PostgresqlColumnAgg, RedshiftColumn): pass

translator = SqlTranslator.from_mappings(
        RedshiftColumn, RedshiftColumnAgg
        )



# Trees and Visitors ==========================================================
from .calls import Call, FuncArg
from .error import ShortException
from .symbolic import strip_symbolic

class FunctionLookupError(ShortException): pass

class FunctionLookupBound:
    def __init__(self, msg):
        self.msg = msg

    def __call__(self, *args, **kwargs):
        raise NotImplementedError(self.msg)


class CallVisitor:
    """
    A node visitor base class that walks the call tree and calls a
    visitor function for every node found.  This function may return a value
    which is forwarded by the `visit` method.

    Note: essentially a copy of ast.NodeVisitor
    """

    def visit(self, node):
        """Visit a node."""
        method = 'visit_' + node.func
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        
        node.map_subcalls(self.visit)

    @classmethod
    def quick_visitor(cls, visit_dict, node):
        """Class method to quickly define and run a custom visitor."""

        qv = type('QuickVisitor', cls, visit_dict)
        qv.visit(node)

class CallListener:
    """Generic listener. Each exit is called on a node's copy."""
    def enter(self, node):
        method = 'enter_' + node.func
        f_enter = getattr(self, method, self.generic_enter)

        return f_enter(node)

    def exit(self, node):
        method = 'exit_' + node.func
        f_exit = getattr(self, method, self.generic_exit)
        return f_exit(node)

    def generic_enter(self, node):
        args, kwargs = node.map_subcalls(self.enter)

        return self.exit(node.__class__(node.func, *args, **kwargs))

    def generic_exit(self, node):
        return node

    def enter_if_call(self, x):
        if isinstance(x, Call):
            return self.enter(x)

        return x
    
    def visit(self, node):
        return self.enter(node)

def get_attr_chain(node, max_n):
    # TODO: need to make custom calls their own Call class, then will not have to
    #       do these kinds of checks, since a __call__ will always be on a Call obj
    if not isinstance(node, Call):
        return [], node

    out = []
    ttl_n = 0
    crnt_node = node
    while ttl_n < max_n:
        if crnt_node.func != "__getattr__": break

        obj, attr = crnt_node.args
        out.append(attr)

        ttl_n += 1
        crnt_node = obj

    return list(reversed(out)), crnt_node


from inspect import isclass, isfunction
from typing import get_type_hints
from .utils import is_dispatch_func_subtype

class CallTreeLocal(CallListener):
    def __init__(
            self,
            local,
            call_sub_attr = None,
            chain_sub_attr = False,
            dispatch_cls = None,
            result_cls = None,
            call_props = None
            ):
        """
        Parameters
        ----------
            local : mapping
                A dictionary mapping func_name: func, used to replace call expressions.
            call_sub_attr : set, optional
                A set of attributes signaling any subattributes are property
                methods. Eg. {'dt'} to signify in _.dt.year, year is a property call.
            chain_sub_attr : bool
                Whether to included the attributes in the above argument, when looking up
                up a replacement for the property call. E.g. does local have a 'dt.year' entry.
            dispatch_cls : class
                If custom calls are dispatchers, dispatch on this class. If none, use their name
                to try and get their corresponding local function.
            result_cls : class 
                If custom calls are dispatchers, require their result annotation to be a subclass
                of this class.
            call_props : set, sequence
                Property methods to potentially convert to local calls.
        """
        self.local = local
        self.call_sub_attr = set(call_sub_attr or [])
        self.chain_sub_attr = chain_sub_attr
        self.dispatch_cls = dispatch_cls
        self.result_cls = result_cls
        self.call_props = set(call_props or [])

    def translate(self, expr):
        """Return the translation of an expression.

        This method is meant to be a high-level entrypoint.

        """

        # note that by contract, don't need to strip symbolic
        return self.enter(strip_symbolic(expr))

    def dispatch_local(self, name):
        f_local = self.local[name]

        if isinstance(f_local, FunctionLookupBound):
            raise FunctionLookupError(f_local.msg)

        return f_local

    def create_local_call(self, name, prev_obj, cls, func_args = None, func_kwargs = None):
        # need call attr name (arg[0].args[1]) 
        # need call arg and kwargs
        func_args = tuple() if func_args is None else func_args
        func_kwargs = {} if func_kwargs is None else func_kwargs

        try:
            local_func = self.dispatch_local(name)
        except KeyError as err:
            raise FunctionLookupError("Missing translation for function call: %s"% name)

        return cls(
                "__call__",
                FuncArg(local_func),
                prev_obj,
                *func_args,
                **func_kwargs
                )

    def enter(self, node):
        # if no enter metthod for operators, like __invert__, try to get from local
        # TODO: want to only do this if func is the name of an infix op's method
        method = 'enter_' + node.func
        if not hasattr(self, method) and node.func in self.local:
            args, kwargs = node.map_subcalls(self.enter)
            return self.create_local_call(node.func, args[0], Call, args[1:], kwargs)

        return super().enter(node)

    def enter___getattr__(self, node):
        obj, attr = node.args

        # Note that the conversion of attribute chains to calls is similar
        # to that in the __call__ visit. But in the code below we're handling
        # optional cases where attribute chains might have property methods,
        # which may (optionally) be converted to calls.
        attr_chain, target = get_attr_chain(node, 2)
        if len(attr_chain) == 2 and attr_chain[0] in self.call_sub_attr:
            # convert subattributes to calls, e.g. dt.days -> dt.days()

            # TODO: should always call exit?
            if self.chain_sub_attr:
                # use chained attribute to look up local function instead
                # e.g. `dt.round`(), rather than `round`()
                attr = ".".join(attr_chain)
            return self.create_local_call(attr, target, Call)
        elif attr in self.call_props:
            return self.create_local_call(attr, obj, Call)

        return self.generic_enter(node)


    def enter___call__(self, node):
        """
        Overview:
             variables      _.x.method(1)         row_number(_.x, 1)
             ---------      -------------        --------------------
            
                            █─'__call__'         █─'__call__'                           
             obj            ├─█─.                ├─<function row_number
                            │ ├─█─.              ├─█─.                                  
                            │ │ ├─_              │ ├─_                                  
                            │ │ └─'x'            │ └─'x'                                
                            │ └─'method'         │                                   
                            └─1                  └─1
        """
        obj, *rest = node.args
        args = tuple(self.enter_if_call(child) for child in rest)
        kwargs = {k: self.enter_if_call(child) for k, child in node.kwargs.items()}

        attr_chain, target = get_attr_chain(obj, max_n = 2)
        if attr_chain:
            # want _.x.method() -> method(_.x), need to transform
            if len(attr_chain) == 2 and attr_chain[0] in self.call_sub_attr:
                # e.g. _.dt.round()
                call_name = ".".join(attr_chain) if self.chain_sub_attr else attr_chain[-1]
                entered_target = self.enter_if_call(target)
            else:
                call_name = attr_chain[-1]
                entered_target = self.enter_if_call(obj.args[0])

        else:
            # default to generic enter
            return self.generic_enter(node)

        return self.create_local_call(
                call_name, entered_target, node.__class__,
                args, kwargs
                )


class ExecutionValidatorVisitor(CallListener):
    # MC-NOTE TOTEST: 
    # * dispatch input parent validates
    # * dispatch output child validates
    def __init__(
            self,
            dispatch_cls = None,
            result_cls = None,
            ):
        """
        Parameters
        ----------
            dispatch_cls : class
                If custom calls are dispatchers, dispatch on this class. If none, use their name
                to try and get their corresponding local function.
            result_cls : class 
                If custom calls are dispatchers, require their result annotation to be a subclass
                of this class.
        """
        self.dispatch_cls = dispatch_cls
        self.result_cls = result_cls

    def validate_dispatcher(self, dispatcher, strict=True):
        f_concrete = dispatcher.dispatch(self.dispatch_cls)
        if isinstance(f_concrete, FunctionLookupBound) and strict:
            raise FunctionLookupError(f_concrete.msg)

        if isclass(f_concrete) and issubclass(f_concrete, Exception):
            raise f_concrete

        return f_concrete

    @staticmethod
    def is_dispatcher(f):
        # TODO: this is essentially a protocol
        return hasattr(f, 'registry') and hasattr(f, 'dispatch')


    def enter___custom_func__(self, node):
        func = node(None)
        
        if self.is_dispatcher(func) and self.dispatch_cls is not None:
            # allow custom functions that dispatch on dispatch_cls
            f = self.validate_dispatcher(func)

            if (self.result_cls is None
                or is_dispatch_func_subtype(f, self.dispatch_cls, self.result_cls)
                ):
                # TODO: MC-NOTE: recreates old behavior, as a temporary step toward codata
                return self.exit(FuncArg(func.dispatch(self.dispatch_cls)))
            
            raise FunctionLookupError(
                    "External function {name} can dispatch on the class {dispatch_cls}, but "
                    "must also have result annotation of (sub)type {result_cls}"
                        .format(
                            name = func.__name__,
                            dispatch_cls = self.dispatch_cls,
                            result_cls = self.result_cls
                            )
                    )

        return self.generic_enter(node)


class CodataVisitor(ExecutionValidatorVisitor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.codata_instance = self.dispatch_cls()

    def exit___call__(self, node):
        if isinstance(node.args[0], FuncArg):
            return Call(
                    "__call__",
                    node.args[0],
                    self.codata_instance,
                    *node.args[1:],
                    **node.kwargs
                    )
        return node





import pandas as pd
import siuba.ops

from sqlalchemy import sql
from sqlalchemy import types as sqla_types

from siuba.dply.verbs import collect, distinct
from siuba.sql import LazyTbl
from siuba.sql.utils import (
    _sql_select,
    _sql_column_collection,
    _sql_add_columns,
    _sql_with_only_columns,
    _sql_case
)

from .pivot_wide import (
    pivot_wider,
    pivot_wider_spec,
    build_wider_spec,
    _is_select_everything,
    _tidy_select,
    _select_expr_slice
)

from .utils import vec_as_names
from .sql_pivot_long import _safe_to_dict

_OPS_DEFAULT=siuba.ops.max

@build_wider_spec.register
def _build_wider_spec(__data: LazyTbl, *args, **kwargs):
    # building the spec only really needs the columns names. however, because we 
    # matched tidyr behavior, we just pass a DataFrame in for now.
    raise NotImplementedError(
        "build_wider_spec currently requires a DataFrame. Please collect() your "
        f"data first. Received type: {type(__data)}"
    )


@pivot_wider_spec.register
def _pivot_wider_spec(
    __data: LazyTbl,
    spec,
    names_repair = "check_unique",
    id_cols = None,
    id_expand = False,
    values_fill = None,
    values_fn = _OPS_DEFAULT,
    unused_fn = None
):
    # Key differences:
    # * values_fn by default is "MAX"

    lazy_tbl = __data
    __data = pd.DataFrame(columns = list(__data.last_op.alias().columns.keys()))

    if id_expand:
        raise NotImplementedError()

    if values_fill is not None:
        raise NotImplementedError()

    if isinstance(values_fn, str):
        _f = lazy_tbl.translator.aggregate.local.get(values_fn)
        if _f is None:
            raise ValueError(
                f"values_fn={repr(values_fn)} does not have a SQL translation."
            )
        values_fn = _f
    elif not hasattr(values_fn, "dispatch"):
        raise NotImplementedError(
            "values_fn currently must be column operation function. For example:\n\n"
            "from siuba.ops import mean\n"
            "pivot_wider(..., values_fn=mean)"
        )

    # TODO: all of this down to "pivot to wide" taken from original func ------
    if _is_select_everything(id_cols):
        id_cols = None

    if unused_fn is not None:
        raise NotImplementedError()

    if not isinstance(id_expand, bool):
        raise TypeError("`id_expand` argument must be True or False.")

    
    # tidyselect ----

    name_vars = spec.columns[~spec.columns.isin([".name", ".value"])].tolist()
    val_vars = spec.loc[:, ".value"].unique().tolist()

    # select id columns
    if id_cols is None:
        others = {*name_vars, *val_vars}
        id_vars = [col for col in __data.columns if col not in others]
    else:
        id_vars = _tidy_select(__data, id_cols, "id_cols")

    id_var_bad = set(id_vars) & set([*name_vars, *val_vars])
    if id_var_bad:
        raise ValueError(
            "id_cols contains columns that are in "
            f"names_from or values_from: {id_var_bad}."
        )


    # pivot to wide -----------------------------------------------------------
    
    # each row of spec becomes a CASE_WHEN.
    # spec columns: .name, .value, <name_cols...>
    # SELECT
    #     FN(CASE
    #         WHEN {<name_col_key1>} == {<name_col_val1>} and [...] THEN {.value1}
    #     ) AS .name1,
    #     ... AS .name2,
    #     ... AS .name3
    sel_alias = lazy_tbl.last_op.alias()
    sel_cols = sel_alias.columns
    dispatch_cls = lazy_tbl.translator.aggregate.dispatch_cls

    wide_name_cols = []
    for row in _safe_to_dict(spec, orient="records"):
        when_clause = sql.and_(sel_cols[k] == row[k] for k in name_vars)
        when_then = (when_clause, sel_cols[row[".value"]])

        col = values_fn(dispatch_cls(), _sql_case([when_then]))

        wide_name_cols.append(col)

    wide_id_cols = [sel_cols[id_] for id_ in id_vars]

    _all_raw_names = list(map(str, [*id_vars, *spec[".name"]]))
    repaired_names = vec_as_names(_all_raw_names, repair=names_repair)
    labeled_cols = [
        col.label(name) for name, col in
        zip(repaired_names, [*wide_id_cols, *wide_name_cols])
    ]
    
    final_sel = _sql_select(labeled_cols).group_by(*wide_id_cols)

    return lazy_tbl.append_op(final_sel)


# simply calls build_wider_spec and pivot_wider_spec
@pivot_wider.register
def _pivot_wider(
    __data: LazyTbl,
    id_cols=None,
    id_expand=False,
    names_from="name",
    names_prefix="",
    names_sep="_",
    names_glue=None,
    names_sort=None,
    names_vary="fastest",
    names_expand=False,
    names_repair="check_unique",
    values_from="value",
    values_fill=None,
    values_fn=_OPS_DEFAULT,
    unused_fn=None
):
    # note that we use three forms of the data: __data for tidyselect,
    # distinct_data for spec creation, and lazy_tbl for the actual pivot
    lazy_tbl = __data    
    __data = pd.DataFrame(columns = list(__data.last_op.alias().columns.keys()))

    # tidyselect variable names -----------------------------------------------
    # adapted from pivot_wide
    name_vars = _tidy_select(__data, names_from, "names_from")
    val_vars = _tidy_select(__data, values_from, "values_from")
    if id_cols is None:
        others = {*name_vars, *val_vars}

        id_cols = tuple([col for col in __data.columns if col not in others])
        id_vars = _select_expr_slice(id_cols)
    else:
        id_vars = id_cols


    # create dummy data with all names_from levels ----------------------------
    distinct_data = collect(distinct(lazy_tbl, *name_vars)).copy()
    distinct_data[list(val_vars)] = True

    vec_as_names(list(distinct_data.columns), repair="check_unique")


    # build spec and pivot ----------------------------------------------------
    spec = build_wider_spec(
        distinct_data,
        names_from = names_from,
        values_from = values_from,
        names_prefix = names_prefix,
        names_sep = names_sep,
        names_glue = names_glue,
        names_sort = names_sort,
        names_vary = names_vary,
        names_expand = names_expand
    )


    out = pivot_wider_spec(
        lazy_tbl,
        spec,
        names_repair = names_repair,
        id_cols = id_vars,
        id_expand = id_expand,
        values_fill = values_fill,
        values_fn = values_fn,
        unused_fn = unused_fn
    )

    return out



from .generics import ALL_OPS, PLAIN_OPS
from .utils import _register_series_default

# register default series methods on all operations
for _generic in ALL_OPS.values():
    _register_series_default(_generic)

del _generic
del _register_series_default


# import accessor generics. These are included in ALL_OPS, but since we want
# users to be able to import from them, also need to be modules. Start their
# names with underscores just to keep the files together.

globals().update(PLAIN_OPS)




# symbolic dispatch wrapper ---------------------------------------------------

from functools import singledispatch, update_wrapper, wraps
import inspect

from .calls import Call, FuncArg, MetaArg, Lazy, PipeCall, _Isolate
from .symbolic import Symbolic, create_sym_call, strip_symbolic

from typing import Callable

def _dispatch_not_impl(func_name):
    def f(x, *args, **kwargs):
        raise TypeError("singledispatch function {func_name} not implemented for type {type}"
                            .format(func_name = func_name, type = type(x))
                            )

    return f

def symbolic_dispatch(f = None, cls = object):
    """Return a generic dispatch function with symbolic data implementations.

    The function dispatches (Call or Symbolic) -> FuncArg.

    Parameters
    ----------
    cls :
        A class to dispatch on.
    f :
        A function to call if no classes match while dispatching.

    Examples
    --------

    Here is an example of running separate add functions on integers and strings.

    >>> @symbolic_dispatch(cls = int)
    ... def add1(x): return x + 1
    
    >>> @add1.register(str)
    ... def _add1_str(x): return int(x) + 1

    >>> add1(1)
    2

    >>> add1("1")
    2

    Note that passing a symbolic causes it to return a symbolic, so you can continue
    creating expressions.

    >>> from siuba.siu import _
    >>> type(add1(_.a.b) + _.c.d)
    <class 'siuba.siu.symbolic.Symbolic'>
    
    symbolic dispatch raises a NotImplementedError by default if it no function ``f``
    is passed. However, you can override the default as follows:
    
    >>> @symbolic_dispatch
    ... def my_func(x): raise NotImplementedError("some error message")

    """
    if f is None:
        return lambda f: symbolic_dispatch(f, cls)

    # TODO: don't use singledispatch if it has already been done
    dispatch_func = singledispatch(f)

    if cls is not object:
        dispatch_func.register(cls, f)
        dispatch_func.register(object, _dispatch_not_impl(dispatch_func.__name__))


    @dispatch_func.register(Symbolic)
    def _dispatch_symbol(__data, *args, **kwargs):
        return create_sym_call(FuncArg(dispatch_func), strip_symbolic(__data), *args, **kwargs)

    @dispatch_func.register(Call)
    def _dispatch_call(__data, *args, **kwargs):
        # TODO: want to just create call, for now use hack of creating a symbolic
        #       call and getting the source off of it...
        return strip_symbolic(create_sym_call(FuncArg(dispatch_func), __data, *args, **kwargs))

    return dispatch_func


# Verb dispatch  ==============================================================

class NoArgs: pass

def pipe_no_args(f):
    """Register a concrete function that handles when a verb received no arguments."""
    @f.register(NoArgs)
    def wrapper(__data, *args, **kwargs):
        # e.g. head() -> Pipeable(_ -> head(_))
        return create_pipe_call(f, MetaArg("_"), *args, **kwargs)

    return f

def register_pipe(f, cls):
    """Register a concrete function that returns a Pipeable when called."""
    @f.register(cls)
    def wrapper(*args, **kwargs):
        return create_pipe_call(f, MetaArg("_"), *args, **kwargs)
    return f

def register_pipe_call(f):
    """Register a concrete function that ."""
    @f.register(Call)
    def f_dispatch(__data, *args, **kwargs):
        call = __data
        if isinstance(call, MetaArg):
            # single _ passed as first arg to function
            # e.g. mutate(_, _.id) -> Pipeable(_ -> mutate(_, _.id))
            return create_pipe_call(f, call, *args, **kwargs)
        else:
            # more complex _ expr passed as first arg to function
            # e.g. mutate(_.id) -> Pipeable(_ -> mutate(_, _.id))
            return create_pipe_call(f, MetaArg("_"), call, *args, **kwargs)

    return f



# option: no args, custom dispatch (e.g. register NoArgs)
# strips symbols
def verb_dispatch(cls, f = None):
    """Wrap singledispatch. Making sure to keep its attributes on the wrapper.
    
    This wrapper has three jobs:
        1. strip symbols off of calls
        2. pass NoArgs instance for calls like some_func(), so dispatcher can handle
        3. return a Pipeable when the first arg of a call is a symbol

    Parameters
    ----------
    cls :
        A class to dispatch on.
    f : 
        A function to call if no classes match while dispatching.
    """

    # classic way of allowing args to a decorator
    if f is None:
        return lambda f: verb_dispatch(cls, f)

    # initially registers func for object, so need to change to pd.DataFrame
    dispatch_func = singledispatch(f)
    if isinstance(cls, tuple):
        for c in cls: dispatch_func.register(c, f)
    else:
        dispatch_func.register(cls, f)
    # then, set the default object dispatcher to create a pipe
    register_pipe(dispatch_func, object)

    # register dispatcher for Call, and NoArgs
    register_pipe_call(dispatch_func)
    pipe_no_args(dispatch_func)

    @wraps(dispatch_func)
    def wrapper(*args, **kwargs):
        strip_args = map(strip_symbolic, args)
        strip_kwargs = {k: strip_symbolic(v) for k,v in kwargs.items()}

        if not args:
            return dispatch_func(NoArgs(), **strip_kwargs)

        return dispatch_func(*strip_args, **strip_kwargs)

    return wrapper

# TODO: deprecate / remove singledispatch2
singledispatch2 = verb_dispatch


# Pipe ========================================================================

class Pipeable:
    """DEPRECATED: please use the siuba.siu.call function.

    Enable function composition through the right bitshift (>>) operator.

    Parameters
    ----------
    f :
        A function to be called.
    calls : sequence, optional
        A list-like of functions to be called, with each result chained into the next.

    Examples
    --------

    >>> f = lambda x: x + 1

    Eager evaluation:

    >>> 1 >> Pipeable(f)
    2

    Defer to a pipe:

    >>> p = Pipeable(f) >> Pipeable(f)
    >>> 1 >> p
    3

    >>> p_undo = p >> (lambda x: x - 3)
    >>> 1 >> p_undo
    0

    >>> from siuba.siu import _
    >>> p_undo_sym = p >> (_ - 3)
    >>> 1 >> p_undo_sym
    0

    """

    def __init__(self, f = None, calls = None):
        import warnings
        warnings.warn("Pipeable is deprecated. Please use siuba.siu.call.")
        # symbolics like _.some_attr need to be stripped down to a call, because
        # calling _.some_attr() returns another symbolic.
        f = strip_symbolic(f)

        if f is not None:
            if calls is not None: raise Exception()
            self.calls = [f]
        elif calls is not None:
            self.calls = calls

    def __rshift__(self, x) -> "Pipeable":
        """Defer evaluation when pipe is on the left (lazy piping)."""
        if isinstance(x, Pipeable):
            return Pipeable(calls = self.calls + x.calls)
        elif isinstance(x, (Symbolic, Call)):
            call = strip_symbolic(x)
            return Pipeable(calls = self.calls + [call])
        elif callable(x):
            return Pipeable(calls = self.calls + [x])

        raise Exception()

    def __rrshift__(self, x):
        """Potentially evaluate result when pipe is on the right (eager piping).

        This function handles two cases:
            * callable >> pipe -> pipe
            * otherwise, evaluate the pipe

        """
        if isinstance(x, (Symbolic, Call)):
            call = strip_symbolic(x)
            return Pipeable(calls = [call] + self.calls)
        elif callable(x):
            return Pipeable(calls = [x] + self.calls)

        return self(x)

    def __call__(self, x):
        """Evaluate a pipe and return the result.

        Parameters
        ----------
        x :
            An object to be passed into the first function in the pipe.

        """
        res = x
        for f in self.calls:
            res = f(res)
        return res

def _prep_isolate_args(*args):
    result = []
    for ii, arg in enumerate(args):
        if ii == 0:
            result.append(strip_symbolic(arg))
        else:
            result.append(_Isolate(strip_symbolic(arg)))

    return result


def create_pipe_call(obj, *args, **kwargs) -> Call:
    """Return a Call of a function on its args and kwargs, wrapped in a Pipeable."""

    stripped_args = _prep_isolate_args(*args)

    return Call(
            "__call__",
            strip_symbolic(obj),
            *stripped_args,
            **{k: _Isolate(strip_symbolic(v)) for k,v in kwargs.items()}
            )

def create_eager_pipe_call(obj, *args, **kwargs) -> Call:

    return Call(
            "__call__",
            strip_symbolic(obj),
            *map(strip_symbolic, args),
            **{k: strip_symbolic(v) for k,v in kwargs.items()}
            )


def call(__func: "Callable | Call | Symbolic", *args, **kwargs):
    """Allow a function call to be used in a call (with >>).

    Parameters
    ----------
    __func:
        A function to be called as part of a call.
    *args:
        Additional position arguments to pass to the function.
    **kwargs:
        Additional keyword arguments to pass to the function.

    Examples
    --------

    The simplest use of the call is passing just the to-be-called function.

    >>> s = "a string"
    >>> s >> call(print)
    a string

    This is equivalent to explicitly passing ``_`` as a placeholder.

    >>> from siuba import _
    >>> s >> call(print, _)
    a string

    The explicit syntax is useful, because it allows us to pass more arguments.
    For example, the code below passes additional arguments to print.

    >>> "a" >> call(print, _, "string", sep=" ")
    a string

    You can transform the input data. For example, the code below passes "shout".upper()
    to print.

    >>> "shout" >> call(print, _.upper())
    SHOUT

    Since ``_`` is just a placeholder for the data on the left-hand-side of >>, you
    can pass it multiple times to the to-be-called function (e.g. print).

    >>> "nice" >> call(print, _, _, sep=" ")
    nice nice

    Alternatively, you can pass a siu expression straight to call.

    >>> "abc" >> call(_[0].upper())
    'A'
        
    """

    if isinstance(__func, (Symbolic, Call)):
        if args or kwargs:
            raise NotImplementedError(
                "If a siu expression (e.g. _) is the first argument to call, it must "
                "be the only argument. You can pass arguments using the form, "
                "call(_.some_method(1, 2, c = 3))."
            )
        return strip_symbolic(__func)
    if not args and not kwargs:
        # handle implicit case, call(some_func) -> call(some_func, _)
        return create_eager_pipe_call(__func, MetaArg("_"))

    return create_eager_pipe_call(__func, *args, **kwargs)


def pipe(__data, *args: Callable):
    """Pipe data through a chain of callables. Return the final result.

    Examples
    --------

    Case 1: pipe regular functions

    >>> pipe({"a": 1}, lambda x: x["a"], lambda x: x + 1)
    2

    Case 2: pipe to siu expressions

    >>> from siuba import _
    >>> pipe({"a": 1}, _["a"], _ + 1)
    2

    Case 3: call external function on siu expression

    >>> from siuba.siu import call
    >>> pipe({"a": 1}, call(isinstance, _["a"], int))
    True

    Case 4: _ as first arg to delay

    >>> f = pipe(_, lambda x: x["a"])
    >>> f
    PipeCall(...)

    >>> f({"a": 1})
    1

    Example: using with verb

    >>> from siuba import _, summarize
    >>> from siuba.data import mtcars
    >>> pipe(mtcars, summarize(res = _.hp.mean()))
            res
    0  146.6875
    """

    stripped = strip_symbolic(__data)

    # Special case: support backwards compatibility with old pipe() behavior ----
    # call() and Call.__rrshift__ now handle this behavior.
    if len(args) == 0:
        if isinstance(stripped, Call):
            return stripped
        else:
            return call(stripped)
        

    # When data is _, return a pipe call
    pipe_call = PipeCall(stripped, *map(strip_symbolic, args))

    if isinstance(stripped, MetaArg):
        return pipe_call

    return pipe_call()



from siuba.dply.across import across, _get_name_template, _across_setup_fns, ctx_verb_data, ctx_verb_window
from siuba.dply.tidyselect import var_select, var_create
from siuba.siu import FormulaContext, Call, FormulaArg
from siuba.siu.calls import str_to_getitem_call
from siuba.siu.visitors import CallListener

from .backend import LazyTbl
from .utils import _sql_select, _sql_column_collection
from .translate import ColumnCollection

from sqlalchemy import sql


class ReplaceFx(CallListener):
    def __init__(self, replacement):
        self.replacement = replacement

    def exit(self, node):
        res = super().exit(node)
        if isinstance(res, FormulaArg):
            return str_to_getitem_call(self.replacement)

        return res


@across.register(LazyTbl)
def _across_lazy_tbl(__data: LazyTbl, cols, fns, names: "str | None" = None) -> LazyTbl:
    raise NotImplementedError(
        "across() cannot called directly on a LazyTbl. Please use it inside a verb, "
        "like mutate(), summarize(), filter(), arrange(), group_by(), etc.."
    )
    #selectable = __data.last_op
    #
    #columns = selectable.alias().columns
    #if not isinstance(columns, ImmutableColumnCollection):
    #    raise TypeError(str(type(columns)))

    #res_cols = across(columns, cols, fns, names)

    #return __data.append_op(_sql_select(res_cols))


@across.register(ColumnCollection)
def _across_sql_cols(
    __data: ColumnCollection,
    cols,
    fns,
    names: "str | None" = None
) -> ColumnCollection:

    lazy_tbl = ctx_verb_data.get()
    window = ctx_verb_window.get()

    column_names = list(__data.keys())

    name_template = _get_name_template(fns, names)
    selected_cols = var_select(column_names, *var_create(cols), data=__data)

    fns_map = _across_setup_fns(fns)
    
    results = []

    # iterate over columns ----
    for new_name, old_name in selected_cols.items():
        if old_name is None:
            old_name = new_name

        crnt_col = __data[old_name]
        #context = FormulaContext(Fx=crnt_col, _=__data)

        # iterate over functions ----
        for fn_name, fn in fns_map.items():
            fmt_pars = {"fn": fn_name, "col": new_name}

            fn_replaced = ReplaceFx(old_name).enter(fn)
            new_call = lazy_tbl.shape_call(
                fn_replaced,
                window,
                verb_name="Across",
                arg_name = f"function {fn_name} of {len(fns_map)}"
            )

            res, windows, _ = lazy_tbl.track_call_windows(new_call, __data)

            #res = new_call(context)
            res_name = name_template.format(**fmt_pars)
            results.append(res.label(res_name))

    return _sql_column_collection(results)



from setuptools import setup, find_packages

# parse version ---------------------------------------------------------------

import re
import ast
_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('siuba/__init__.py', 'rb') as f:
    VERSION = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

with open('README.md', encoding="utf-8") as f:
    README = f.read()

# setup -----------------------------------------------------------------------

setup(
    name='siuba',
    packages=find_packages(),
    version=VERSION,
    description='A package for quick, scrappy analyses with pandas and SQL',
    author='Michael Chow',
    license='MIT',
    author_email='mc_al_gh_siuba@fastmail.com',
    url='https://github.com/machow/siuba',
    keywords=['package', ],
    install_requires=[
        "pandas>=0.24.0,<2.1.0",
        "numpy>=1.12.0",
        "SQLAlchemy>=1.2.19",
        "PyYAML>=3.0.0"
    ],
    extras_require={
        "test": [
            "pytest",
            "hypothesis",
            "IPython",
            "pymysql",
            "psycopg2-binary",
            "duckdb_engine",
            # duckdb 0.8.0 has a bug which always errors for pandas v2+
            # it's been fixed, but we need to pin until duckdb v0.9.0
            "duckdb",
        ],
        "docs": [
            "plotnine",
            "jupyter",
            "nbval",
            "sphinx",
            "nbsphinx",
            "jupytext",
            "gapminder==0.1",
        ],
    },
    python_requires=">=3.7",
    include_package_data=True,
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)




import numpy as np
import pytest

from siuba.siu import strip_symbolic, FunctionLookupError, Symbolic, MetaArg, Call, _ as D


# Note that currently tests are split across the test_siu.py, and this module.

@pytest.fixture
def _():
    return Symbolic()

def test_siu_symbolic_np_array_ufunc_call(_):
    sym = np.add(_, 1)
    expr = strip_symbolic(sym)

    # structure:
    # █─'__call__'
    # ├─█─'__custom_func__'
    # │ └─<function array_ufunc at 0x103aa3820>
    # ├─_
    # ├─<ufunc 'add'>
    # ├─'__call__'
    # ├─_
    # └─1

    assert len(expr.args) == 6
    assert expr.args[1] is strip_symbolic(_)        # original dispatch obj
    assert expr.args[2] is np.add                   # ufunc object
    assert expr.args[3] == "__call__"               # its method to use
    assert expr.args[4] is strip_symbolic(_)        # lhs input
    assert expr.args[5] == 1                        # rhs input


def test_siu_symbolic_np_array_ufunc_inputs_lhs(_):
    lhs = np.array([1,2])
    rhs = np.array([3,4])
    res = lhs + rhs

    # symbol on lhs ----

    sym = np.add(_, rhs)
    expr = strip_symbolic(sym)

    assert np.array_equal(expr(lhs), res)


def test_siu_symbolic_np_array_ufunc_inputs_rhs(_):
    lhs = np.array([1,2])
    rhs = np.array([3,4])
    res = lhs + rhs

    # symbol on rhs ----

    sym2 = np.add(lhs, _)
    expr2 = strip_symbolic(sym2)

    assert np.array_equal(expr2(rhs), res)


@pytest.mark.xfail
def test_siu_symbolic_np_array_function(_):
    # Note that np.sum is not a ufunc, but sort of reduces on a ufunc under the
    # hood, so fails when called on a symbol
    sym = np.sum(_)
    expr = strip_symbolic(sym)

    assert expr(np.array([1,2])) == 3


@pytest.mark.parametrize("func", [
    np.absolute,   # a ufunc
    np.sum         # dispatched by __array_function__
    ])
def test_siu_symbolic_array_ufunc_sql_raises(_, func):
    from siuba.sql.utils import mock_sqlalchemy_engine
    from siuba.sql import LazyTbl
    from siuba.sql import SqlFunctionLookupError

    lazy_tbl = LazyTbl(mock_sqlalchemy_engine("postgresql"), "somedata", ["x", "y"])
    with pytest.raises(SqlFunctionLookupError) as exc_info:
        lazy_tbl.shape_call(strip_symbolic(func(_.x)))

    assert "Numpy function sql translation" in exc_info.value.args[0]
    assert "not supported" in exc_info.value.args[0]



@pytest.mark.parametrize("sym, res", [
    (np.sqrt(D), lambda ser: np.sqrt(ser)),    # ufunc
    (np.add(D, 1), lambda ser: np.add(ser, 1)),
    (np.add(1, D), lambda ser : np.add(1, ser)),
    (np.add(D, D), lambda ser : np.add(ser, ser)),
])
def test_siu_symbolic_array_ufunc_pandas(_, sym, res):
    import pandas as pd
    ser = pd.Series([1,2])

    expr = strip_symbolic(sym)

    src = expr(ser)
    dst = res(ser)

    assert isinstance(src, pd.Series)
    assert src.equals(dst)


@pytest.mark.parametrize("sym, res", [
    (np.mean(D), lambda ser : np.mean(ser)),     # __array_function__
    (np.sum(D), lambda ser: np.sum(ser)), 
    (np.sqrt(np.mean(D)), lambda ser: np.sqrt(np.mean(ser))), 
])
def test_siu_symbolic_array_function_pandas(_, sym, res):
    import pandas as pd
    ser = pd.Series([1,2])

    expr = strip_symbolic(sym)

    src = expr(ser)
    dst = res(ser)

    # note that all examples currently are aggregates
    assert src == dst



"""
Tests the singledispatch functions in siuba.dply.vector. Note that these functions
also include doctests, so we want to mostly test their implementations return similar
results.
"""
import pytest

from siuba import _, arrange, mutate, group_by, summarize, filter
import siuba.sql.dply
from siuba.dply import vector as v
from datetime import timedelta

from hypothesis.strategies import text, floats, integers
from hypothesis.extra.pandas import data_frames, column, indexes 

from .helpers import assert_equal_query, data_frame, backend_sql
from pandas.testing import assert_frame_equal

DATA_SPEC = data_frames([
        column('x', elements = floats(width = 32) | integers(), unique = True),
        column('g', dtype = str, elements = text(max_size = 1))
        ],
        index = indexes(elements = floats() | integers(), max_size = 10)
    )

OMNIBUS_VECTOR_FUNCS = [
        #cumall, cumany, cummean,
        #desc,
        ## removed below, since pandas was giving 0 and eg .1^(-16) same rank
        #v.dense_rank(_.x, na_option = "keep"),
        #v.percent_rank(_.x),
        v.min_rank(_.x, na_option = "keep"),
        v.cume_dist(_.x, na_option = "keep"),
        v.row_number(_.x),
        #ntile,
        v.between(_.x, 2, 5, default = False),
        v.coalesce(_.x, 2),
        v.lead(_.x),
        v.lag(_.x),
        v.n(_.x),
        v.na_if(_.x, 2),
        #near,
        v.nth(_.x, 2),
        v.first(_.x),
        v.last(_.x, order_by = _.idx),            # TODO: in SQL getting FROM LAST requires order by
        ]

VECTOR_AGG_FUNCS = [
        v.n(_.x),
        v.n(_),
        ]

VECTOR_FILTER_FUNCS = [
        v.dense_rank(_.x, na_option = "keep") < 2,
        v.min_rank(_.x, na_option = "keep") < 2,
        v.cume_dist(_.x, na_option = "keep") < .3,
        v.row_number(_.x) < 2,
        v.between(_.x, 2, 3),
        v.lead(_.x) > _.x,
        v.lag(_.x) > _.x,
        v.n(_.x) == 1,
        v.na_if(_.x, 1).isna()
        ]

@pytest.fixture(params = [
    data_frame(idx = [1,2,3], x = [1,2,3], g = ['a', 'a', 'b']),
    data_frame(idx = [1,2,3], x = [1.,2.,3.], g = ['a', 'a', 'b']),
    data_frame(idx = [1,2,3], x = [1.,2.,None], g = ['a', 'a', 'b']),
    ])
def simple_data(request):
    return request.param

@pytest.mark.parametrize('func', OMNIBUS_VECTOR_FUNCS)
def test_mutate_vector(backend, func, simple_data):
    df = backend.load_cached_df(simple_data)
    
    assert_equal_query(
            df,
            arrange(_.idx) >> mutate(y = func),
            simple_data.assign(y = func),
            check_dtype = False
            )

    # grouped
    assert_equal_query(
            df,
            arrange(_.idx) >> group_by(_.g) >> mutate(y = func),
            simple_data.groupby('g').apply(lambda d: d.assign(y = func)).reset_index(drop = True),
            check_dtype = False
            )


@pytest.mark.parametrize('func', VECTOR_AGG_FUNCS)
def test_agg_vector(backend, func, simple_data):
    # This test uses pandas .apply method, which no longer works with
    # symbolic expressions. This was unsafe to do anyway. Now, we
    # explicitly convert it to a call.
    from siuba.siu import strip_symbolic

    df = backend.load_cached_df(simple_data)

    call = strip_symbolic(func)

    res = data_frame(y = call(simple_data))

    assert_equal_query(
            df,
            summarize(y = func),
            res 
            )

    # grouped
    assert_equal_query(
            df,
            group_by(_.g) >> summarize(y = func),
            simple_data.groupby('g').apply(call).reset_index().rename(columns = {0: 'y'})
            )


@backend_sql
@pytest.mark.parametrize('func', VECTOR_FILTER_FUNCS)
def test_filter_vector(backend, func, simple_data):
    df = backend.load_cached_df(simple_data)

    res = data_frame(y = func(simple_data))

    assert_equal_query(
            df,
            arrange(_.idx) >> filter(func),
            filter(simple_data, func),
            # ignore dtypes, since sql -> an empty data frame has object columns
            check_dtype = False
            )

    # grouped (vs slow_filter)
    assert_equal_query(
            df,
            arrange(_.idx) >> group_by(_.g) >> filter(func),
            simple_data >> group_by(_.g) >> filter(func),
            check_dtype = False
            )



import pandas as pd
import numpy as np

from siuba.siu import symbolic_dispatch
from collections import defaultdict


def _get_cat_order(x):
    if isinstance(x, pd.Series):
        arr = x.array
    else:
        arr = x

    if isinstance(arr, pd.Categorical):
        return arr.ordered

    return None

def _maybe_upcast(fct_in, fct_out):
    if isinstance(fct_in, pd.Series):
        return pd.Series(fct_out)

    return fct_out


# fct_inorder, fct_infreq -----------------------------------------------------

@symbolic_dispatch
def fct_inorder(fct, ordered=None):
    """Return a copy of fct, with categories ordered by when they first appear.

    Parameters
    ----------
    fct : list-like
        A pandas Series, Categorical, or list-like object
    ordered : bool
        Whether to return an ordered categorical. By default a Categorical inputs'
        ordered setting is respected. Use this to override it.

    See Also
    --------
    fct_infreq : Order categories by value frequency count.

    Examples
    --------
    
    >>> fct = pd.Categorical(["c", "a", "b"])
    >>> fct
    ['c', 'a', 'b']
    Categories (3, object): ['a', 'b', 'c']

    Note that above the categories are sorted alphabetically. Use fct_inorder
    to keep the categories in first-observed order.

    >>> fct_inorder(fct)
    ['c', 'a', 'b']
    Categories (3, object): ['c', 'a', 'b']

    fct_inorder also accepts pd.Series and list objects:

    >>> fct_inorder(["z", "a"])
    ['z', 'a']
    Categories (2, object): ['z', 'a']

    By default, the ordered setting of categoricals is respected. Use the ordered
    parameter to override it.

    >>> fct2 = pd.Categorical(["z", "a", "b"], ordered=True)
    >>> fct_inorder(fct2)
    ['z', 'a', 'b']
    Categories (3, object): ['z' < 'a' < 'b']

    >>> fct_inorder(fct2, ordered=False)
    ['z', 'a', 'b']
    Categories (3, object): ['z', 'a', 'b']

    """

    if ordered is None:
        ordered = _get_cat_order(fct)

    if isinstance(fct, (pd.Series, pd.Categorical)):
        uniq = fct.dropna().unique()

        if isinstance(uniq, pd.Categorical):
            # the result of .unique for a categorical is a new categorical
            # unsurprisingly, it also sorts the categories, so reorder manually
            # (note that this also applies to Series[Categorical].unique())
            categories = uniq.categories[uniq.dropna().codes]
            return pd.Categorical(fct, categories, ordered=ordered)

        # series in, so series out
        cat = pd.Categorical(fct, uniq, ordered=ordered)
        return pd.Series(cat)

    ser = pd.Series(fct)
    return pd.Categorical(fct, categories = ser.dropna().unique(), ordered=ordered)


@symbolic_dispatch
def fct_infreq(fct, ordered=None):
    """Return a copy of fct, with categories ordered by frequency (largest first)

    Parameters
    ----------
    fct : list-like
        A pandas Series, Categorical, or list-like object
    ordered : bool
        Whether to return an ordered categorical. By default a Categorical inputs'
        ordered setting is respected. Use this to override it.

    See Also
    --------
    fct_inorder : Order categories by when they're first observed.

    Examples
    --------

    >>> fct_infreq(["c", "a", "c", "c", "a", "b"])
    ['c', 'a', 'c', 'c', 'a', 'b']
    Categories (3, object): ['c', 'a', 'b']

    """

    if ordered is None:
        ordered = _get_cat_order(fct)


    # sort and create new categorical ----
    
    if isinstance(fct, pd.Categorical):
        # Categorical value counts are sorted in categories order
        # So to acheive the exact same result as the Series case below,
        # we need to use fct_inorder, so categories is in first-observed order.
        # This orders the final result by frequency, and then observed for ties.
        freq = fct_inorder(fct).value_counts().sort_values(ascending=False)

        # note that freq is a Series, but it has a CategoricalIndex.
        # we want the index values as shown, so we need to strip them out of
        # this nightmare index situation.
        categories = freq.index.categories[freq.index.dropna().codes]
        return pd.Categorical(fct, categories=categories, ordered=ordered)

    else:
        # Series sorts in descending frequency order
        ser = pd.Series(fct) if not isinstance(fct, pd.Series) else fct
        freq = ser.value_counts()
        cat = pd.Categorical(ser, categories=freq.index, ordered=ordered)

        if isinstance(fct, pd.Series):
            return pd.Series(cat)

        return cat


# fct_reorder -----------------------------------------------------------------

@symbolic_dispatch
def fct_reorder(fct, x, func = np.median, desc = False) -> pd.Categorical:
    """Return copy of fct, with categories reordered according to values in x.
    
    Parameters
    ----------
    fct :
        A pandas.Categorical, or array(-like) used to create one.
    x :
        Values used to reorder categorical. Must be same length as fct.
    func :
        Function run over all values within a level of the categorical.
    desc :
        Whether to sort in descending order.

    Notes
    -----
    NaN categories can't be ordered. When func returns NaN, sorting
    is always done with NaNs last.


    Examples
    --------

    >>> fct_reorder(['a', 'a', 'b'], [4, 3, 2])
    ['a', 'a', 'b']
    Categories (2, object): ['b', 'a']

    >>> fct_reorder(['a', 'a', 'b'], [4, 3, 2], desc = True)
    ['a', 'a', 'b']
    Categories (2, object): ['a', 'b']

    >>> fct_reorder(['x', 'x', 'y'], [4, 0, 2], np.max)
    ['x', 'x', 'y']
    Categories (2, object): ['y', 'x']

    """

    x_vals = x.values if isinstance(x, pd.Series) else x
    s = pd.Series(x_vals, index = fct)

    # sort groups by calculated agg func. note that groupby uses dropna=True by default,
    # but that's okay, since pandas categoricals can't order the NA category
    ordered = s.groupby(level = 0).agg(func).sort_values(ascending = not desc)

    out = pd.Categorical(fct, categories=ordered.index)
    return _maybe_upcast(fct, out)


# fct_recode ------------------------------------------------------------------

@symbolic_dispatch
def fct_recode(fct, recat=None, **kwargs) -> pd.Categorical:
    """Return copy of fct with renamed categories.

    Parameters
    ----------
    fct :
        A pandas.Categorical, or array(-like) used to create one. 
    **kwargs :
        Arguments of form new_name = old_name.

    Examples
    --------
    >>> cat = ['a', 'b', 'c']
    >>> fct_recode(cat, z = 'c')
    ['a', 'b', 'z']
    Categories (3, object): ['a', 'b', 'z']

    >>> fct_recode(cat, x = ['a', 'b'])
    ['x', 'x', 'c']
    Categories (2, object): ['x', 'c']

    >>> fct_recode(cat, {"x": ['a', 'b']})
    ['x', 'x', 'c']
    Categories (2, object): ['x', 'c']
        
    """

    if recat and not isinstance(recat, dict):
        raise TypeError("fct_recode requires named args or a dict.")

    if recat and kwargs:
        duplicate_keys = set(recat).intersection(set(kwargs))
        if duplicate_keys:
            raise ValueError(
                "The following recode name(s) were specified more than once: {}" \
                .format(duplicate_keys)
            )

    new_cats = {**recat, **kwargs} if recat else kwargs
    return _maybe_upcast(fct, fct_collapse(fct, new_cats))


# fct_collapse ----------------------------------------------------------------

@symbolic_dispatch
def fct_collapse(fct, recat, group_other = None) -> pd.Categorical:
    """Return copy of fct with categories renamed. Optionally group all others.

    Parameters
    ----------
    fct :
        A pandas.Categorical, or array(-like) used to create one.  
    recat :
        Dictionary of form {new_cat_name: old_cat_name}. old_cat_name may be
        a list of existing categories, to be given the same name.
    group_other :
        An optional string, specifying what all other categories should be named.
        This will always be the last category level in the result.

    Notes
    -----
    Resulting levels index is ordered according to the earliest level replaced.
    If we rename the first and last levels to "c", then "c" is the first level.

    Examples
    --------
    >>> fct_collapse(['a', 'b', 'c'], {'x': 'a'})
    ['x', 'b', 'c']
    Categories (3, object): ['x', 'b', 'c']

    >>> fct_collapse(['a', 'b', 'c'], {'x': 'a'}, group_other = 'others')
    ['x', 'others', 'others']
    Categories (2, object): ['x', 'others']

    >>> fct_collapse(['a', 'b', 'c'], {'ab': ['a', 'b']})
    ['ab', 'ab', 'c']
    Categories (2, object): ['ab', 'c']

    >>> fct_collapse(['a', 'b', None], {'a': ['b']})
    ['a', 'a', NaN]
    Categories (1, object): ['a']

    """
    if not isinstance(fct, pd.Categorical):
        new_fct = pd.Categorical(fct)
    else:
        new_fct = fct

    # each existing cat will map to a new one ----
    # need to know existing to new cat
    # need to know new cat to new code
    cat_to_new = {k: None for k in new_fct.categories}
    for new_name, v in recat.items():
        v = [v] if not np.ndim(v) else v
        for old_name in v:
            if cat_to_new[old_name] is not None:
                raise Exception("category %s was already re-assigned"%old_name)
            cat_to_new[old_name] = new_name

    # collapse all unspecified cats to group_other if specified ----
    for k, v in cat_to_new.items():
        if v is None:
            if group_other is not None:
                cat_to_new[k] = group_other
            else:
                cat_to_new[k] = k

    # map from old cat to new code ----

    # calculate new codes
    ordered_cats = {new: True for old, new in cat_to_new.items()}

    # move the other group to last in the ordered set
    if group_other is not None:
        try:
            del ordered_cats[group_other]
            ordered_cats[group_other] = True
        except KeyError:
            pass

    # map new category name to code
    new_cat_set = {k: ii for ii, k in enumerate(ordered_cats)}

    # at this point, we need remap codes to the other category
    # make an array, where the index is old code + 1 (so missing val index is 0)
    old_code_to_new = np.array(
            [-1] + [new_cat_set[new_cat] for new_cat in cat_to_new.values()]
    )

    # map old cats to new codes
    #remap_code = {old: new_cat_set[new] for old, new in cat_to_new.items()}
    new_codes = old_code_to_new[new_fct.codes + 1]
    new_cats = list(new_cat_set)

    out = pd.Categorical.from_codes(new_codes, new_cats)
    return _maybe_upcast(fct, out)


# fct_lump --------------------------------------------------------------------

@symbolic_dispatch
def fct_lump(fct, n = None, prop = None, w = None, other_level = "Other", ties = None) -> pd.Categorical:
    """Return a copy of fct with categories lumped together.

    Parameters
    ----------
    fct :
        A pandas.Categorical, or array(-like) used to create one.
    n :
        Number of categories to keep.
    prop :
        (not implemented) keep categories that occur prop proportion of the time.
    w :
        Array of weights corresponding to each value in fct.
    other_level :
        Name for all lumped together levels.
    ties :
        (not implemented) method to use in the case of ties.

    Notes
    -----
    Currently, one of n and prop must be specified.

    Examples
    --------
    >>> fct_lump(['a', 'a', 'b', 'c'], n = 1)
    ['a', 'a', 'Other', 'Other']
    Categories (2, object): ['a', 'Other']

    >>> fct_lump(['a', 'a', 'b', 'b', 'c', 'd'], prop = .2)
    ['a', 'a', 'b', 'b', 'Other', 'Other']
    Categories (3, object): ['a', 'b', 'Other']
        
    """

    if ties is not None:
        raise NotImplementedError("ties is not implemented")

    if n is None and prop is None:
        raise NotImplementedError("Either n or prop must be specified")

    keep_cats = _fct_lump_n_cats(fct, w, other_level, ties, n = n, prop = prop)

    out = fct_collapse(fct, {k:k for k in keep_cats}, group_other = other_level)
    return _maybe_upcast(fct, out)


def _fct_lump_n_cats(fct, w, other_level, ties, n = None, prop = None):
    # TODO: currently always selects n, even if ties

    # weights might be a Series, or array, etc..
    arr = _get_values(w) if w is not None else 1
    ser = pd.Series(arr, index = fct)
    counts = ser.groupby(level = 0).sum()

    if n is not None:
        ascending = n < 0
        sorted_arr = counts.sort_values(ascending = ascending)
        res = sorted_arr.iloc[:abs(n)]
    elif prop is not None:
        sorted_arr = counts.sort_values() / counts.sum()

        if prop < 0:
            res = sorted_arr.loc[sorted_arr <= abs(prop)]
        else:
            res = sorted_arr.loc[sorted_arr > prop]

    return res.index.values

def _get_values(x):
    # TODO: move into utility, note pandas now encouraging .array method
    if isinstance(x, pd.Series): return x.values

    return x


# fct_rev ---------------------------------------------------------------------

@symbolic_dispatch
def fct_rev(fct) -> pd.Categorical:
    """Return a copy of fct with category level order reversed.next
    
    Parameters
    ----------
    fct :
        A pandas.Categorical, or array(-like) used to create one.

    Examples
    --------
    >>> fct = pd.Categorical(["a", "b", "c"])
    >>> fct
    ['a', 'b', 'c']
    Categories (3, object): ['a', 'b', 'c']

    >>> fct_rev(fct)
    ['a', 'b', 'c']
    Categories (3, object): ['c', 'b', 'a']

    Note that this function can also accept a list.

    >>> fct_rev(["a", "b", "c"])
    ['a', 'b', 'c']
    Categories (3, object): ['c', 'b', 'a']


    """

    if not isinstance(fct, pd.Categorical):
        fct = pd.Categorical(fct)

    rev_levels = list(reversed(fct.categories))

    out = fct.reorder_categories(rev_levels)
    return _maybe_upcast(fct, out)



from siuba.dply.verbs import collect

from ..backend import LazyTbl
from ..utils import _FixedSqlDatabase, _is_dialect_duckdb, MockConnection

# collect ----------

@collect.register(LazyTbl)
def _collect(__data, as_df = True):
    # TODO: maybe remove as_df options, always return dataframe

    if isinstance(__data.source, MockConnection):
        # a mock sqlalchemy is being used to show_query, and echo queries.
        # it doesn't return a result object or have a context handler, so
        # we need to bail out early
        return

    # compile query ----

    if _is_dialect_duckdb(__data.source):
        # TODO: can be removed once next release of duckdb fixes:
        # https://github.com/duckdb/duckdb/issues/2972
        query = __data.last_select
        compiled = query.compile(
            dialect = __data.source.dialect,
            compile_kwargs = {"literal_binds": True}
        )
    else:
        compiled = __data.last_select

    # execute query ----

    with __data.source.connect() as conn:
        if as_df:
            sql_db = _FixedSqlDatabase(conn)

            if _is_dialect_duckdb(__data.source):
                # TODO: pandas read_sql is very slow with duckdb.
                # see https://github.com/pandas-dev/pandas/issues/45678
                # going to handle here for now. address once LazyTbl gets
                # subclassed per backend.
                duckdb_con = conn.connection.c
                return duckdb_con.query(str(compiled)).to_df()
            else:
                #
                return sql_db.read_sql(compiled)

        return conn.execute(compiled)
